@@ -7,6 +7,7 @@
 #include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/wait.h>
 #include <time.h>
 
@@ -61,8 +62,10 @@ static int check_processes (time_t timeout)
     
     current = process_head;
     while (current != NULL) {
-        if (now - current->time > timeout)
+        if (now - current->time > timeout) {
+            remove_process (current->pid);
             return (ETOOLONG);
+        }
         current = current->next;
     }
     return (ENOERR);
@@ -85,7 +88,6 @@ int check_bin(char *tbinary, time_t timeout)
 
     child_pid = fork();
     if (!child_pid) {
-	
 	/* child, exit immediately, if no test binary given */
 	if (tbinary == NULL)
 	    exit(0);
@@ -93,12 +95,16 @@ int check_bin(char *tbinary, time_t timeout)
 	/* Don't want the stdin and stdout of our test program
 	 * to cause trouble
 	 * So make stdout and stderr go to their respective files */	
-	if (!freopen("/var/log/watchdog/test-bin.stdout", "a+", stdout))
+	strcpy(filename_buf, logdir);
+	strcat(filename_buf, "/test-bin.stdout");
+	if (!freopen(filename_buf, "a+", stdout))
 	    exit (errno);
-	if (!freopen("/var/log/watchdog/test-bin.stderr", "a+", stderr))
+	strcpy(filename_buf, logdir);
+	strcat(filename_buf, "/test-bin.stderr");
+	if (!freopen(filename_buf, "a+", stderr))
 	    exit (errno);
-	
-	/* else start binary */
+
+	/* now start binary */
 	execl(tbinary, tbinary, NULL);
 
 	/* execl should only return in case of an error */
@@ -124,7 +130,7 @@ int check_bin(char *tbinary, time_t timeout)
 
 	/* wait for child(s) to stop */
 	/* but only after a short sleep */
-	sleep(tint >> 1);
+	usleep(tint * 500000);
 
 	do {
 	    ret = waitpid(-1, &result, WNOHANG);
@@ -155,7 +161,7 @@ int check_bin(char *tbinary, time_t timeout)
 	} else {
 		/* in case there are still old childs running due to an error */
 		/* log that error */
-		if (err != 0 && err != ECHILD) {
+		if (ret != 0 && err != 0 && err != ECHILD) {
 #if USE_SYSLOG
 		    errno = err;
 		    syslog(LOG_ERR, "child %d did not exit immediately (error = %d = '%m')", child_pid, err);
