@@ -16,6 +16,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
+#include <limits.h>
 #include <netdb.h>
 #include <sched.h>
 #include <signal.h>
@@ -29,7 +30,6 @@
 #include <linux/watchdog.h>
 #include <string.h>
 
-#include <sys/ioctl.h>
 #include <libgen.h>
 
 #include <unistd.h>
@@ -40,6 +40,8 @@
 #include <syslog.h>
 #endif				/* USE_SYSLOG */
 
+#define max(x,y) (x) > (y) ? (x) : (y)
+
 static int no_act = FALSE;
 
 #if USE_SYSLOG
@@ -67,23 +69,27 @@ volatile sig_atomic_t _running = 1;
 #define PRIORITY	"priority"
 #define REALTIME	"realtime"
 #define REPAIRBIN	"repair-binary"
+#define REPAIRTIMEOUT	"repair-timeout"
 #define TEMP		"temperature-device"
 #define TESTBIN		"test-binary"
 #define TESTTIMEOUT	"test-timeout"
 #define HEARTBEAT	"heartbeat-file"
 #define HBSTAMPS	"heartbeat-stamps"
+#define LOGDIR		"log-dir"
 
 pid_t pid;
-int softboot = FALSE, watchdog = -1, load = -1, mem = -1, temp = -1;
-int tint = 10, logtick = 1, ticker = 1, schedprio = 1;
+int tint = 1, softboot = FALSE, watchdog = -1, load = -1, mem = -1, temp = -1;
+int logtick = 1, ticker = 1, schedprio = 1;
 int maxload1 = 0, maxload5 = 0, maxload15 = 0, minpages = 0;
 int maxtemp = 120, hbstamps = 300, lastts, nrts;
 int pingcount = 3;
 int devtimeout = TIMER_MARGIN;
 char *tempname = NULL, *devname = NULL, *admin = "root", *progname;
 char *timestamps, *heartbeat;
-time_t timeout = 0;
+time_t timeout = 0, rtimeout = 0;
 FILE *hb;
+char* logdir = "/var/log/watchdog";
+char *filename_buf;
 
 #if defined(_POSIX_MEMLOCK)
 int mlocked = FALSE, realtime = FALSE;
@@ -114,6 +120,7 @@ static int sync_system(int sync_it)
 static int repair(char *rbinary, int result, char *name)
 {
     pid_t child_pid;
+    pid_t r_pid;
     char parm[5];
     int ret;
 
@@ -125,15 +132,19 @@ static int repair(char *rbinary, int result, char *name)
 
     child_pid = fork();
     if (!child_pid) {
-	/* Don't want the stdin and stdout of our test program
-	 * to cause trouble
+	/* Don't want the stdin and stdout of our repair program
+	 * to cause trouble.
 	 * So make stdout and stderr go to their respective files */
-        if (!freopen("/var/log/watchdog/repair-bin.stdout", "a+", stdout))
-        	exit (errno);
-        if (!freopen("/var/log/watchdog/repair-bin.stderr", "a+", stderr))
-        	exit (errno);
+	strcpy(filename_buf, logdir);
+	strcat(filename_buf, "/repair-bin.stdout");
+	if (!freopen(filename_buf, "a+", stdout))
+		exit (errno);
+	strcpy(filename_buf, logdir);
+	strcat(filename_buf, "/repair-bin.stderr");
+	if (!freopen(filename_buf, "a+", stderr))
+		exit (errno);
 
-        /* else start binary */
+        /* now start binary */
 	if (name == NULL)
 		execl(rbinary, rbinary, parm, NULL);
 	else
@@ -155,7 +166,27 @@ static int repair(char *rbinary, int result, char *name)
 	else
 	    return (ENOERR);
     }
-    if (waitpid(child_pid, &result, 0) != child_pid) {
+
+    if (rtimeout > 0) {
+	time_t left = rtimeout;
+	do {
+	    sleep (1);
+	    r_pid = waitpid(child_pid, &result, WNOHANG);
+	    if (r_pid)
+		break;
+	    left--;
+	} while (left > 0);
+
+    } else
+	r_pid = waitpid(child_pid, &result, 0);
+    if (r_pid == 0) {
+#if USE_SYSLOG
+	syslog(LOG_ERR, "repair child %d timed out", child_pid);
+#else				/* USE_SYSLOG */
+	perror(progname);
+#endif				/* USE_SYSLOG */
+	return (EREBOOT);
+    } else if (r_pid != child_pid) {
 	int err = errno;
 
 #if USE_SYSLOG
@@ -241,13 +272,13 @@ static int spool(char *line, int *i, int offset)
     	return(0);
 }
 
-static void read_config(char *filename, char *progname)
+static void read_config(char *configfile, char *progname)
 {
     FILE *wc;
     int gotload5 = FALSE, gotload15 = FALSE;
 
-    if ((wc = fopen(filename, "r")) == NULL) {
-	fprintf(stderr, "%s: Can't open config file \"%s\": %s ", progname, filename, strerror(errno));
+    if ((wc = fopen(configfile, "r")) == NULL) {
+	fprintf(stderr, "%s: Can't open config file \"%s\": %s ", progname, configfile, strerror(errno));
 	exit(1);
     }
 
@@ -335,6 +366,11 @@ static void read_config(char *filename, char *progname)
 			rbinary = NULL;
 		else
 			rbinary = strdup(line + i);
+	    } else if (strncmp(line + i, REPAIRTIMEOUT, strlen(REPAIRTIMEOUT)) == 0) {
+		if (spool(line, &i, strlen(REPAIRTIMEOUT)))
+			rtimeout = 0;
+		else
+			rtimeout = atol(line + i);
 	    } else if (strncmp(line + i, TESTBIN, strlen(TESTBIN)) == 0) {
 		if (spool(line, &i, strlen(TESTBIN)))
 			tbinary = NULL;
@@ -417,9 +453,13 @@ static void read_config(char *filename, char *progname)
 	    } else if (strncmp(line + i, MINMEM, strlen(MINMEM)) == 0) {
 		if (spool(line, &i, strlen(MINMEM)))
 			fprintf(stderr, "Ignoring invalid line in config file:\n%s\n", line);
-		else {
+		else
 			minpages = atol(line + i);
-		}
+	    } else if (strncmp(line + i, LOGDIR, strlen(LOGDIR)) == 0) {
+		if (spool(line, &i, strlen(LOGDIR)))
+			fprintf(stderr, "Ignoring invalid line in config file:\n%s\n", line);
+		else
+			logdir = strdup(line + i);
 	    } else {
 		fprintf(stderr, "Ignoring invalid line in config file:\n%s\n", line);
 	    }
@@ -432,9 +472,9 @@ static void read_config(char *filename, char *progname)
     }
 }
 
-static void old_option(int c, char *filename)
+static void old_option(int c, char *configfile)
 {
-    fprintf(stderr, "Option -%c is no longer valid, please specify it in %s.\n", c, filename);
+    fprintf(stderr, "Option -%c is no longer valid, please specify it in %s.\n", c, configfile);
     usage();
 }
 
@@ -443,7 +483,7 @@ int main(int argc, char *const argv[])
     FILE *fp;
     int c, force = FALSE, sync_it = FALSE;
     int hold;
-    char *filename = CONFIG_FILENAME;
+    char *configfile = CONFIG_FILENAME;
     struct list *act;
     pid_t child_pid;
 
@@ -460,6 +500,7 @@ int main(int argc, char *const argv[])
 	{NULL, 0, NULL, 0}
     };
     long count = 0L;
+    struct watchdog_info ident;
 #else				/* USE_SYSLOG */
     char *opts = "d:i:n:fsbql:p:t:c:r:m:a:";
     struct option long_options[] =
@@ -491,10 +532,10 @@ int main(int argc, char *const argv[])
 	case 'l':
 	case 'm':
 	case 'i':
-	    old_option(c, filename);
+	    old_option(c, configfile);
 	    break;
 	case 'c':
-	    filename = optarg;
+	    configfile = optarg;
 	    break;
 	case 'f':
 	    force = TRUE;
@@ -518,7 +559,7 @@ int main(int argc, char *const argv[])
 	}
     }
 
-    read_config(filename, progname);
+    read_config(configfile, progname);
 
     if (tint < 0)
 	usage();
@@ -537,10 +578,10 @@ int main(int argc, char *const argv[])
 	exit(1);
     }
    
-    /* make sure we get our own directory in /var/log */
-    if (mkdir ("/var/log/watchdog", 0750) && errno != EEXIST) {
+    /* make sure we get our own log directory */
+    if (mkdir (logdir, 0750) && errno != EEXIST) {
 	fprintf(stderr, "%s error:\n", progname);
-        fprintf(stderr, "Cannot create directory /var/log/watchdog\n");
+        fprintf(stderr, "Cannot create directory %s\n", logdir);
 	exit (1);
     }
 
@@ -586,12 +627,21 @@ int main(int argc, char *const argv[])
 	    act->parameter.net = *net;
 	}
     }
-    
+
     /* make sure we're on the root partition */
     if (chdir("/") < 0) {
 	perror(progname);
 	exit(1);
     }
+
+    /* allocate some memory to store a filename, this is needed later on even
+     * if the system runs out of memory */
+    filename_buf = (char*)malloc(max(strlen(logdir) + sizeof("/repair-bin.stdout") + 1, strlen("/proc//oom_adj") + sizeof(int) * CHAR_BIT * 10 / 3 + 1));
+    if (!filename_buf) {
+	error(progname);
+        exit(1);
+    }
+
 #if !defined(DEBUG)
     /* fork to go into the background */
     if ((child_pid = fork()) < 0) {
@@ -663,9 +713,9 @@ int main(int argc, char *const argv[])
             for (act = iface; act != NULL; act = act->next)
                 syslog(LOG_INFO, "interface: %s", act->name);                
 
-    syslog(LOG_INFO, "test=%s(%ld) repair=%s alive=%s heartbeat=%s temp=%s to=%s no_act=%s",
+    syslog(LOG_INFO, "test=%s(%ld) repair=%s(%d) alive=%s heartbeat=%s temp=%s to=%s no_act=%s",
 	    (tbinary == NULL) ? "none" : tbinary, timeout, 
-	    (rbinary == NULL) ? "none" : rbinary,
+	    (rbinary == NULL) ? "none" : rbinary, rtimeout,
 	    (devname == NULL) ? "none" : devname,
 	    (heartbeat == NULL) ? "none" : heartbeat,
 	    (tempname == NULL) ? "none" : tempname,
@@ -673,7 +723,6 @@ int main(int argc, char *const argv[])
 	    (no_act == TRUE) ? "yes" : "no");
 #endif				/* USE_SYSLOG */
 
-
     /* open the device */
     if (devname != NULL && no_act == FALSE) {
 	watchdog = open(devname, O_WRONLY);
@@ -686,16 +735,29 @@ int main(int argc, char *const argv[])
 	    /* do not exit here per default */
 	    /* we can use watchdog even if there is no watchdog device */
 	}
-	if (watchdog >= 0 && devtimeout > 0) {
-	    /* Set the watchdog hard-stop timeout; default = unset (use
-	       driver default) */
-	    if (ioctl(watchdog, WDIOC_SETTIMEOUT, &devtimeout) < 0) {
+	if (watchdog >= 0) {
+	    if (devtimeout > 0) {
+		/* Set the watchdog hard-stop timeout; default = unset (use
+		   driver default) */
+		if (ioctl(watchdog, WDIOC_SETTIMEOUT, &devtimeout) < 0) {
 #if USE_SYSLOG
-                syslog(LOG_ERR, "cannot set timeout %d (errno = %d = '%m')", devtimeout, errno);
+		syslog(LOG_ERR, "cannot set timeout %d (errno = %d = '%m')", devtimeout, errno);
 #else				
-            	perror(progname);
+		perror(progname);
 #endif			   
+		}
 	    }
+
+#if USE_SYSLOG
+	    /* Also log watchdog identity */
+	    if (ioctl(watchdog, WDIOC_GETSUPPORT, &ident) < 0) {
+		syslog(LOG_ERR, "cannot get watchdog identity (errno = %d = '%m')", errno);
+	    }
+	    else {
+		ident.identity[sizeof(ident.identity) - 1] = '\0'; /* Be sure */
+                syslog(LOG_INFO, "hardware wartchdog identity: %s", ident.identity);
+	    }
+#endif
 	}
     }
 
@@ -792,6 +854,7 @@ int main(int argc, char *const argv[])
 	fprintf(fp, "%d\n", pid);
 	(void) fclose(fp);
     }
+
     /* set signal term to set our run flag to 0 so that */
     /* we make sure watchdog device is closed when receiving SIGTERM */
     signal(SIGTERM, sigterm_handler);
@@ -822,6 +885,14 @@ int main(int argc, char *const argv[])
     }
 #endif
 
+    /* tell oom killer to not kill this process */
+    sprintf(filename_buf, "/proc/%d/oom_adj", pid);
+    fp = fopen(filename_buf, "w");
+    if (fp != NULL) {
+	fprintf(fp, "-17\n");
+	(void) fclose(fp);
+    }
+
     /* main loop: update after <tint> seconds */
     while (_running) {
 	wd_action(keep_alive(), rbinary, NULL);
@@ -861,7 +932,7 @@ int main(int argc, char *const argv[])
 	do_check(check_bin(tbinary, timeout), rbinary, NULL);
 
 	/* finally sleep some seconds */
-	sleep((tint >> 1) + (tint % 2)); /* this should make watchdog sleep tint seconds alltogther */
+	usleep(tint * 500000); /* this should make watchdog sleep tint seconds alltogther */
 	/* sleep(tint); */
 
 #if USE_SYSLOG
