@@ -18,12 +18,14 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
+#include <limits.h>
 #include <sched.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/mman.h>
 #include <sys/wait.h>
+#include <linux/watchdog.h>
 #include <libgen.h>
 #include <string.h>
 #include <syslog.h>
@@ -57,8 +59,6 @@ static void usage(void)
     exit(1);
 }
 
-
-
 /* write a log entry on exit */
 static void log_end()
 {
@@ -128,11 +128,11 @@ static int spool(char *line, int *i, int offset)
         return(0);
 }
 
-static void read_config(char *filename, char *progname)
+static void read_config(char *configfile, char *progname)
 {
     FILE *wc;
 
-    if ( (wc = fopen(filename, "r")) == NULL ) {
+    if ( (wc = fopen(configfile, "r")) == NULL ) {
         perror(progname);
         exit(1);
     }
@@ -210,8 +210,8 @@ static void read_config(char *filename, char *progname)
 int main(int argc, char *const argv[])
 {
     FILE *fp;
-    char log[256];
-    char *filename = CONFIG_FILENAME;
+    char *configfile = CONFIG_FILENAME;
+    char *filename_buf;
     pid_t child_pid;
     int count = 0;
     int c;
@@ -228,6 +228,7 @@ int main(int argc, char *const argv[])
 	{"softboot", no_argument, NULL, 'b'},
 	{NULL, 0, NULL, 0}
     };
+    struct watchdog_info ident;
 #else				/* USE_SYSLOG */
     char *opts = "d:i:n:fsbql:p:t:c:r:m:a:";
     struct option long_options[] =
@@ -249,7 +250,7 @@ int main(int argc, char *const argv[])
 	    break;
 	switch (c) {
 	case 'c':
-	    filename = optarg;
+	    configfile = optarg;
 	    break;
 	case 'n':
 	case 'p':
@@ -273,13 +274,22 @@ int main(int argc, char *const argv[])
 	}
     }
 
-    read_config(filename, progname);
+    read_config(configfile, progname);
 
     /* make sure we're on the root partition */
     if ( chdir("/") < 0 ) {
         perror(progname);
         exit(1);
     }
+
+    /* allocate some memory to store a filename, this is needed later on even
+     * if the system runs out of memory */
+    filename_buf = (char*)malloc(strlen("/proc//oom_adj") + sizeof(int) * CHAR_BIT * 10 / 3 + 1);
+    if (!filename_buf) {
+        error(progname);
+        exit(1);
+    }
+
 #if !defined(DEBUG)
     /* fork to go into the background */
     if ( (child_pid = fork()) < 0 ) {
@@ -320,9 +330,8 @@ int main(int argc, char *const argv[])
 
     /* Log the starting message */
     openlog(progname, LOG_PID, LOG_DAEMON);
-    sprintf(log, "starting watchdog keepalive daemon (%d.%d):", MAJOR_VERSION, MINOR_VERSION);
-    sprintf(log + strlen(log), " int=%d alive=%s realtime=%s", tint, devname, realtime ? "yes" : "no");
-    syslog(LOG_INFO, log);
+    syslog(LOG_INFO, "starting watchdog keepalive daemon (%d.%d):", MAJOR_VERSION, MINOR_VERSION);
+    syslog(LOG_INFO, " int=%d alive=%s realtime=%s", tint, devname, realtime ? "yes" : "no");
 #endif                          /* USE_SYSLOG */
 
     /* this daemon has no other function than writing to this device 
@@ -342,6 +351,17 @@ int main(int argc, char *const argv[])
             exit(1);
     }
 
+#if USE_SYSLOG
+    /* Also log watchdog identity */
+    if (ioctl(watchdog, WDIOC_GETSUPPORT, &ident) < 0) {
+	syslog(LOG_ERR, "cannot get watchdog identity (errno = %d = '%m')", errno);
+    }
+    else {
+	ident.identity[sizeof(ident.identity) - 1] = '\0'; /* Be sure */
+	syslog(LOG_INFO, "hardware wartchdog identity: %s", ident.identity);
+    }
+#endif
+
     /* tuck my process id away */
     fp = fopen(KA_PIDFILE, "w");
     if ( fp != NULL ) {
@@ -382,6 +402,14 @@ int main(int argc, char *const argv[])
     }
 #endif
 
+    /* tell oom killer to not kill this process */
+    sprintf(filename_buf, "/proc/%d/oom_adj", getpid());
+    fp = fopen(filename_buf, "w");
+    if (fp != NULL) {
+        fprintf(fp, "-17\n");
+        (void) fclose(fp);
+    }
+
     /* main loop: update after <tint> seconds */
     while ( _running ) {
         if ( write(watchdog, "\0", 1) < 0 ) {
